@@ -10,7 +10,7 @@ from getpass import getpass
 from queue import Empty
 
 import zmq.asyncio
-from traitlets import Any  # type: ignore
+from traitlets import Any
 from traitlets import Bool
 from traitlets import Instance
 from traitlets import Type
@@ -93,7 +93,7 @@ class KernelClient(ConnectionFileMixin):
     # The PyZMQ Context to use for communication with the kernel.
     context = Instance(zmq.asyncio.Context)
 
-    _created_context: Bool = Bool(False)
+    _created_context = Bool(False)
 
     def _context_default(self) -> zmq.asyncio.Context:
         self._created_context = True
@@ -116,23 +116,40 @@ class KernelClient(ConnectionFileMixin):
     # flag for whether execute requests should be allowed to call raw_input:
     allow_stdin: bool = True
 
+    def __del__(self):
+        """Handle garbage collection.  Destroy context if applicable."""
+        if self._created_context and self.context and not self.context.closed:
+            if self.channels_running:
+                if self.log:
+                    self.log.warning("Could not destroy zmq context for %s", self)
+            else:
+                if self.log:
+                    self.log.debug("Destroying zmq context for %s", self)
+                self.context.destroy()
+        try:
+            super_del = super().__del__
+        except AttributeError:
+            pass
+        else:
+            super_del()
+
     # --------------------------------------------------------------------------
     # Channel proxy methods
     # --------------------------------------------------------------------------
 
-    async def _async_get_shell_msg(self, *args, **kwargs) -> t.Dict[str, t.Any]:
+    async def _async_get_shell_msg(self, *args: Any, **kwargs: Any) -> t.Dict[str, t.Any]:
         """Get a message from the shell channel"""
         return await self.shell_channel.get_msg(*args, **kwargs)
 
-    async def _async_get_iopub_msg(self, *args, **kwargs) -> t.Dict[str, t.Any]:
+    async def _async_get_iopub_msg(self, *args: Any, **kwargs: Any) -> t.Dict[str, t.Any]:
         """Get a message from the iopub channel"""
         return await self.iopub_channel.get_msg(*args, **kwargs)
 
-    async def _async_get_stdin_msg(self, *args, **kwargs) -> t.Dict[str, t.Any]:
+    async def _async_get_stdin_msg(self, *args: Any, **kwargs: Any) -> t.Dict[str, t.Any]:
         """Get a message from the stdin channel"""
         return await self.stdin_channel.get_msg(*args, **kwargs)
 
-    async def _async_get_control_msg(self, *args, **kwargs) -> t.Dict[str, t.Any]:
+    async def _async_get_control_msg(self, *args: Any, **kwargs: Any) -> t.Dict[str, t.Any]:
         """Get a message from the control channel"""
         return await self.control_channel.get_msg(*args, **kwargs)
 
@@ -253,7 +270,7 @@ class KernelClient(ConnectionFileMixin):
         self,
         session: Session,
         socket: zmq.sugar.socket.Socket,
-        parent_header,
+        parent_header: Any,
         msg: t.Dict[str, t.Any],
     ) -> None:
         """Output hook when running inside an IPython kernel
@@ -286,9 +303,6 @@ class KernelClient(ConnectionFileMixin):
         :meth:`start_kernel`. If the channels have been stopped and you
         call this, :class:`RuntimeError` will be raised.
         """
-        # Create the context if needed.
-        if not self._created_context:
-            self.context = self._context_default()
         if iopub:
             self.iopub_channel.start()
         if shell:
@@ -318,9 +332,6 @@ class KernelClient(ConnectionFileMixin):
             self.hb_channel.stop()
         if self.control_channel.is_alive():
             self.control_channel.stop()
-        if self._created_context:
-            self._created_context = False
-            self.context.destroy()
 
     @property
     def channels_running(self) -> bool:
@@ -676,7 +687,7 @@ class KernelClient(ConnectionFileMixin):
         raw: bool = True,
         output: bool = False,
         hist_access_type: str = "range",
-        **kwargs,
+        **kwargs: Any,
     ) -> str:
         """Get entries from the kernel's history list.
 
