@@ -8,68 +8,82 @@ Also defined here are utilities for working with Sessions:
 Sessions.
 * A Message object for convenience that allows attribute-access to the msg dict.
 """
-
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
-
-from binascii import b2a_hex
 import hashlib
 import hmac
+import json
 import logging
 import os
 import pickle
 import pprint
 import random
+import typing as t
 import warnings
+from binascii import b2a_hex
 from datetime import datetime
-
-PICKLE_PROTOCOL = pickle.DEFAULT_PROTOCOL
-
-# We are using compare_digest to limit the surface of timing attacks
-from hmac import compare_digest
-
 from datetime import timezone
-utc = timezone.utc
+from hmac import (
+    compare_digest,
+)  # We are using compare_digest to limit the surface of timing attacks
+from typing import Optional
+from typing import Union
 
 import zmq
-from zmq.utils import jsonapi
+from traitlets import Any
+from traitlets import Bool
+from traitlets import CBytes
+from traitlets import CUnicode
+from traitlets import Dict
+from traitlets import DottedObjectName
+from traitlets import Instance
+from traitlets import Integer
+from traitlets import observe
+from traitlets import Set
+from traitlets import TraitError
+from traitlets import Unicode
+from traitlets.config.configurable import Configurable
+from traitlets.config.configurable import LoggingConfigurable
+from traitlets.log import get_logger
+from traitlets.utils.importstring import import_item
 from zmq.eventloop.ioloop import IOLoop
 from zmq.eventloop.zmqstream import ZMQStream
 
-from traitlets.config.configurable import Configurable, LoggingConfigurable
-from ipython_genutils.importstring import import_item
-from jupyter_client.jsonutil import extract_dates, squash_dates, date_default
-from ipython_genutils.py3compat import str_to_bytes, str_to_unicode
-from traitlets import (
-    CBytes, Unicode, Bool, Any, Instance, Set, DottedObjectName, CUnicode,
-    Dict, Integer, TraitError, observe
-)
 from jupyter_client import protocol_version
 from jupyter_client.adapter import adapt
-from traitlets.log import get_logger
+from jupyter_client.jsonutil import extract_dates
+from jupyter_client.jsonutil import json_clean
+from jupyter_client.jsonutil import json_default
+from jupyter_client.jsonutil import squash_dates
 
 
-#-----------------------------------------------------------------------------
+PICKLE_PROTOCOL = pickle.DEFAULT_PROTOCOL
+
+utc = timezone.utc
+
+# -----------------------------------------------------------------------------
 # utility functions
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
+
 
 def squash_unicode(obj):
     """coerce unicode back to bytestrings."""
-    if isinstance(obj,dict):
+    if isinstance(obj, dict):
         for key in obj.keys():
             obj[key] = squash_unicode(obj[key])
             if isinstance(key, str):
                 obj[squash_unicode(key)] = obj.pop(key)
     elif isinstance(obj, list):
-        for i,v in enumerate(obj):
+        for i, v in enumerate(obj):
             obj[i] = squash_unicode(v)
     elif isinstance(obj, str):
-        obj = obj.encode('utf8')
+        obj = obj.encode("utf8")
     return obj
 
-#-----------------------------------------------------------------------------
+
+# -----------------------------------------------------------------------------
 # globals and defaults
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
 
 # default values for the thresholds:
 MAX_ITEMS = 64
@@ -78,12 +92,45 @@ MAX_BYTES = 1024
 # ISO8601-ify datetime objects
 # allow unicode
 # disallow nan, because it's not actually valid JSON
-json_packer = lambda obj: jsonapi.dumps(obj, default=date_default,
-    ensure_ascii=False, allow_nan=False,
-)
-json_unpacker = lambda s: jsonapi.loads(s)
 
-pickle_packer = lambda o: pickle.dumps(squash_dates(o), PICKLE_PROTOCOL)
+
+def json_packer(obj):
+    try:
+        return json.dumps(
+            obj,
+            default=json_default,
+            ensure_ascii=False,
+            allow_nan=False,
+        ).encode("utf8")
+    except (TypeError, ValueError) as e:
+        # Fallback to trying to clean the json before serializing
+        packed = json.dumps(
+            json_clean(obj),
+            default=json_default,
+            ensure_ascii=False,
+            allow_nan=False,
+        ).encode("utf8")
+
+        warnings.warn(
+            f"Message serialization failed with:\n{e}\n"
+            "Supporting this message is deprecated in jupyter-client 7, please make "
+            "sure your message is JSON-compliant",
+            stacklevel=2,
+        )
+
+        return packed
+
+
+def json_unpacker(s):
+    if isinstance(s, bytes):
+        s = s.decode("utf8", "replace")
+    return json.loads(s)
+
+
+def pickle_packer(o):
+    return pickle.dumps(squash_dates(o), PICKLE_PROTOCOL)
+
+
 pickle_unpacker = pickle.loads
 
 default_packer = json_packer
@@ -93,11 +140,12 @@ DELIM = b"<IDS|MSG>"
 # singleton dummy tracker, which will always report as done
 DONE = zmq.MessageTracker()
 
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
 # Mixin tools for apps that use Sessions
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
 
-def new_id():
+
+def new_id() -> str:
     """Generate a new random id.
 
     Avoids problematic runtime import in stdlib uuid on Python 2.
@@ -108,71 +156,79 @@ def new_id():
     id string (16 random bytes as hex-encoded text, chunks separated by '-')
     """
     buf = os.urandom(16)
-    return '-'.join(b2a_hex(x).decode('ascii') for x in (
-        buf[:4], buf[4:]
-    ))
+    return "-".join(b2a_hex(x).decode("ascii") for x in (buf[:4], buf[4:]))
 
-def new_id_bytes():
+
+def new_id_bytes() -> bytes:
     """Return new_id as ascii bytes"""
-    return new_id().encode('ascii')
+    return new_id().encode("ascii")
+
 
 session_aliases = dict(
-    ident = 'Session.session',
-    user = 'Session.username',
-    keyfile = 'Session.keyfile',
+    ident="Session.session",
+    user="Session.username",
+    keyfile="Session.keyfile",
 )
 
-session_flags  = {
-    'secure' : ({'Session' : { 'key' : new_id_bytes(),
-                            'keyfile' : '' }},
+session_flags = {
+    "secure": (
+        {"Session": {"key": new_id_bytes(), "keyfile": ""}},
         """Use HMAC digests for authentication of messages.
         Setting this flag will generate a new UUID to use as the HMAC key.
-        """),
-    'no-secure' : ({'Session' : { 'key' : b'', 'keyfile' : '' }},
-        """Don't authenticate messages."""),
+        """,
+    ),
+    "no-secure": (
+        {"Session": {"key": b"", "keyfile": ""}},
+        """Don't authenticate messages.""",
+    ),
 }
 
-def default_secure(cfg):
+
+def default_secure(cfg: t.Any) -> None:
     """Set the default behavior for a config environment to be secure.
 
     If Session.key/keyfile have not been set, set Session.key to
     a new random UUID.
     """
     warnings.warn("default_secure is deprecated", DeprecationWarning)
-    if 'Session' in cfg:
-        if 'key' in cfg.Session or 'keyfile' in cfg.Session:
+    if "Session" in cfg:
+        if "key" in cfg.Session or "keyfile" in cfg.Session:
             return
     # key/keyfile not specified, generate new UUID:
     cfg.Session.key = new_id_bytes()
 
-def utcnow():
+
+def utcnow() -> datetime:
     """Return timezone-aware UTC timestamp"""
     return datetime.utcnow().replace(tzinfo=utc)
 
-#-----------------------------------------------------------------------------
+
+# -----------------------------------------------------------------------------
 # Classes
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
+
 
 class SessionFactory(LoggingConfigurable):
     """The Base class for configurables that have a Session, Context, logger,
     and IOLoop.
     """
 
-    logname = Unicode('')
+    logname = Unicode("")
 
-    @observe('logname')
-    def _logname_changed(self, change):
-        self.log = logging.getLogger(change['new'])
+    @observe("logname")  # type:ignore[misc]
+    def _logname_changed(self, change: Any) -> None:
+        self.log = logging.getLogger(change["new"])
 
     # not configurable:
-    context = Instance('zmq.Context')
-    def _context_default(self):
+    context = Instance("zmq.Context")
+
+    def _context_default(self) -> zmq.Context:
         return zmq.Context()
 
-    session = Instance('jupyter_client.session.Session',
-                       allow_none=True)
+    session = Instance("jupyter_client.session.Session", allow_none=True)
+
+    loop = Instance("tornado.ioloop.IOLoop")
 
-    loop = Instance('tornado.ioloop.IOLoop')
     def _loop_default(self):
         return IOLoop.current()
 
@@ -190,7 +246,7 @@ class Message(object):
     A Message can be created from a dict and a dict from a Message instance
     simply by calling dict(msg_obj)."""
 
-    def __init__(self, msg_dict):
+    def __init__(self, msg_dict: t.Dict[str, t.Any]) -> None:
         dct = self.__dict__
         for k, v in dict(msg_dict).items():
             if isinstance(v, dict):
@@ -198,39 +254,40 @@ class Message(object):
             dct[k] = v
 
     # Having this iterator lets dict(msg_obj) work out of the box.
-    def __iter__(self):
+    def __iter__(self) -> t.ItemsView[str, t.Any]:
         return self.__dict__.items()
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return repr(self.__dict__)
 
-    def __str__(self):
+    def __str__(self) -> str:
         return pprint.pformat(self.__dict__)
 
-    def __contains__(self, k):
+    def __contains__(self, k: object) -> bool:
         return k in self.__dict__
 
-    def __getitem__(self, k):
+    def __getitem__(self, k: str) -> t.Any:
         return self.__dict__[k]
 
 
-def msg_header(msg_id, msg_type, username, session):
+def msg_header(msg_id: str, msg_type: str, username: str, session: "Session") -> t.Dict[str, t.Any]:
     """Create a new message header"""
     date = utcnow()
     version = protocol_version
     return locals()
 
-def extract_header(msg_or_header):
+
+def extract_header(msg_or_header: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
     """Given a message or header, return the header."""
     if not msg_or_header:
         return {}
     try:
         # See if msg_or_header is the entire message.
-        h = msg_or_header['header']
+        h = msg_or_header["header"]
     except KeyError:
         try:
             # See if msg_or_header is just the header
-            h = msg_or_header['msg_id']
+            h = msg_or_header["msg_id"]
         except KeyError:
             raise
         else:
@@ -239,6 +296,7 @@ def extract_header(msg_or_header):
         h = dict(h)
     return h
 
+
 class Session(Configurable):
     """Object for handling serialization and sending of messages.
 
@@ -282,127 +340,146 @@ class Session(Configurable):
 
     debug = Bool(False, config=True, help="""Debug output in the Session""")
 
-    check_pid = Bool(True, config=True,
+    check_pid = Bool(
+        True,
+        config=True,
         help="""Whether to check PID to protect against calls after fork.
 
         This check can be disabled if fork-safety is handled elsewhere.
-        """)
+        """,
+    )
 
-    packer = DottedObjectName('json',config=True,
-            help="""The name of the packer for serializing messages.
+    packer = DottedObjectName(
+        "json",
+        config=True,
+        help="""The name of the packer for serializing messages.
             Should be one of 'json', 'pickle', or an import name
-            for a custom callable serializer.""")
+            for a custom callable serializer.""",
+    )
 
-    @observe('packer')
+    @observe("packer")
     def _packer_changed(self, change):
-        new = change['new']
-        if new.lower() == 'json':
+        new = change["new"]
+        if new.lower() == "json":
             self.pack = json_packer
             self.unpack = json_unpacker
             self.unpacker = new
-        elif new.lower() == 'pickle':
+        elif new.lower() == "pickle":
             self.pack = pickle_packer
             self.unpack = pickle_unpacker
             self.unpacker = new
         else:
             self.pack = import_item(str(new))
 
-    unpacker = DottedObjectName('json', config=True,
+    unpacker = DottedObjectName(
+        "json",
+        config=True,
         help="""The name of the unpacker for unserializing messages.
-        Only used with custom functions for `packer`.""")
+        Only used with custom functions for `packer`.""",
+    )
 
-    @observe('unpacker')
+    @observe("unpacker")
     def _unpacker_changed(self, change):
-        new = change['new']
-        if new.lower() == 'json':
+        new = change["new"]
+        if new.lower() == "json":
             self.pack = json_packer
             self.unpack = json_unpacker
             self.packer = new
-        elif new.lower() == 'pickle':
+        elif new.lower() == "pickle":
             self.pack = pickle_packer
             self.unpack = pickle_unpacker
             self.packer = new
         else:
             self.unpack = import_item(str(new))
 
-    session = CUnicode('', config=True,
-        help="""The UUID identifying this session.""")
-    def _session_default(self):
+    session = CUnicode("", config=True, help="""The UUID identifying this session.""")
+
+    def _session_default(self) -> str:
         u = new_id()
-        self.bsession = u.encode('ascii')
+        self.bsession = u.encode("ascii")
         return u
 
-    @observe('session')
+    @observe("session")
     def _session_changed(self, change):
-        self.bsession = self.session.encode('ascii')
+        self.bsession = self.session.encode("ascii")
 
     # bsession is the session as bytes
-    bsession = CBytes(b'')
+    bsession = CBytes(b"")
 
-    username = Unicode(str_to_unicode(os.environ.get('USER', 'username')),
+    username = Unicode(
+        os.environ.get("USER", "username"),
         help="""Username for the Session. Default is your system username.""",
-        config=True)
+        config=True,
+    )
 
-    metadata = Dict({}, config=True,
-        help="""Metadata dictionary, which serves as the default top-level metadata dict for each message.""")
+    metadata = Dict(
+        {},
+        config=True,
+        help="Metadata dictionary, which serves as the default top-level metadata dict for each "
+        "message.",
+    )
 
     # if 0, no adapting to do.
     adapt_version = Integer(0)
 
     # message signature related traits:
 
-    key = CBytes(config=True,
-        help="""execution key, for signing messages.""")
-    def _key_default(self):
+    key = CBytes(config=True, help="""execution key, for signing messages.""")
+
+    def _key_default(self) -> bytes:
         return new_id_bytes()
 
-    @observe('key')
+    @observe("key")
     def _key_changed(self, change):
         self._new_auth()
 
-    signature_scheme = Unicode('hmac-sha256', config=True,
+    signature_scheme = Unicode(
+        "hmac-sha256",
+        config=True,
         help="""The digest scheme used to construct the message signatures.
-        Must have the form 'hmac-HASH'.""")
+        Must have the form 'hmac-HASH'.""",
+    )
 
-    @observe('signature_scheme')
+    @observe("signature_scheme")
     def _signature_scheme_changed(self, change):
-        new = change['new']
-        if not new.startswith('hmac-'):
+        new = change["new"]
+        if not new.startswith("hmac-"):
             raise TraitError("signature_scheme must start with 'hmac-', got %r" % new)
-        hash_name = new.split('-', 1)[1]
+        hash_name = new.split("-", 1)[1]
         try:
             self.digest_mod = getattr(hashlib, hash_name)
         except AttributeError as e:
-            raise TraitError("hashlib has no such attribute: %s" %
-                             hash_name) from e
+            raise TraitError("hashlib has no such attribute: %s" % hash_name) from e
         self._new_auth()
 
     digest_mod = Any()
-    def _digest_mod_default(self):
+
+    def _digest_mod_default(self) -> t.Callable:
         return hashlib.sha256
 
     auth = Instance(hmac.HMAC, allow_none=True)
 
-    def _new_auth(self):
+    def _new_auth(self) -> None:
         if self.key:
             self.auth = hmac.HMAC(self.key, digestmod=self.digest_mod)
         else:
             self.auth = None
 
     digest_history = Set()
-    digest_history_size = Integer(2**16, config=True,
+    digest_history_size = Integer(
+        2**16,
+        config=True,
         help="""The maximum number of digests to remember.
 
         The digest history will be culled when it exceeds this value.
-        """
+        """,
     )
 
-    keyfile = Unicode('', config=True,
-        help="""path to file containing execution key.""")
+    keyfile = Unicode("", config=True, help="""path to file containing execution key.""")
 
-    @observe('keyfile')
+    @observe("keyfile")
     def _keyfile_changed(self, change):
-        with open(change['new'], 'rb') as f:
+        with open(change["new"], "rb") as f:
             self.key = f.read().strip()
 
     # for protecting against sends from forks
@@ -410,35 +487,43 @@ class Session(Configurable):
 
     # serialization traits:
 
-    pack = Any(default_packer) # the actual packer function
+    pack = Any(default_packer)  # the actual packer function
 
-    @observe('pack')
+    @observe("pack")
     def _pack_changed(self, change):
-        new = change['new']
+        new = change["new"]
         if not callable(new):
-            raise TypeError("packer must be callable, not %s"%type(new))
+            raise TypeError("packer must be callable, not %s" % type(new))
 
-    unpack = Any(default_unpacker) # the actual packer function
+    unpack = Any(default_unpacker)  # the actual packer function
 
-    @observe('unpack')
+    @observe("unpack")
     def _unpack_changed(self, change):
         # unpacker is not checked - it is assumed to be
-        new = change['new']
+        new = change["new"]
         if not callable(new):
-            raise TypeError("unpacker must be callable, not %s"%type(new))
+            raise TypeError("unpacker must be callable, not %s" % type(new))
 
     # thresholds:
-    copy_threshold = Integer(2**16, config=True,
-        help="Threshold (in bytes) beyond which a buffer should be sent without copying.")
-    buffer_threshold = Integer(MAX_BYTES, config=True,
-        help="Threshold (in bytes) beyond which an object's buffer should be extracted to avoid pickling.")
-    item_threshold = Integer(MAX_ITEMS, config=True,
+    copy_threshold = Integer(
+        2**16,
+        config=True,
+        help="Threshold (in bytes) beyond which a buffer should be sent without copying.",
+    )
+    buffer_threshold = Integer(
+        MAX_BYTES,
+        config=True,
+        help="Threshold (in bytes) beyond which an object's buffer should be extracted to avoid "
+        "pickling.",
+    )
+    item_threshold = Integer(
+        MAX_ITEMS,
+        config=True,
         help="""The maximum number of items for a container to be introspected for custom serialization.
         Containers larger than this are pickled outright.
-        """
+        """,
     )
 
-
     def __init__(self, **kwargs):
         """create a Session object
 
@@ -487,9 +572,11 @@ class Session(Configurable):
         self.pid = os.getpid()
         self._new_auth()
         if not self.key:
-            get_logger().warning("Message signing is disabled.  This is insecure and not recommended!")
+            get_logger().warning(
+                "Message signing is disabled.  This is insecure and not recommended!"
+            )
 
-    def clone(self):
+    def clone(self) -> "Session":
         """Create a copy of this Session
 
         Useful when connecting multiple times to a given kernel.
@@ -508,63 +595,62 @@ class Session(Configurable):
         return new_session
 
     message_count = 0
+
     @property
-    def msg_id(self):
+    def msg_id(self) -> str:
         message_number = self.message_count
         self.message_count += 1
-        return '{}_{}'.format(self.session, message_number)
+        return f"{self.session}_{os.getpid()}_{message_number}"
 
-    def _check_packers(self):
+    def _check_packers(self) -> None:
         """check packers for datetime support."""
         pack = self.pack
         unpack = self.unpack
 
         # check simple serialization
-        msg = dict(a=[1,'hi'])
+        msg_list = dict(a=[1, "hi"])
         try:
-            packed = pack(msg)
+            packed = pack(msg_list)
         except Exception as e:
-            msg = "packer '{packer}' could not serialize a simple message: {e}{jsonmsg}"
-            if self.packer == 'json':
-                jsonmsg = "\nzmq.utils.jsonapi.jsonmod = %s" % jsonapi.jsonmod
-            else:
-                jsonmsg = ""
             raise ValueError(
-                msg.format(packer=self.packer, e=e, jsonmsg=jsonmsg)
+                f"packer '{self.packer}' could not serialize a simple message: {e}"
             ) from e
 
         # ensure packed message is bytes
         if not isinstance(packed, bytes):
-            raise ValueError("message packed to %r, but bytes are required"%type(packed))
+            raise ValueError("message packed to %r, but bytes are required" % type(packed))
 
         # check that unpack is pack's inverse
         try:
             unpacked = unpack(packed)
-            assert unpacked == msg
+            assert unpacked == msg_list
         except Exception as e:
-            msg = "unpacker '{unpacker}' could not handle output from packer '{packer}': {e}{jsonmsg}"
-            if self.packer == 'json':
-                jsonmsg = "\nzmq.utils.jsonapi.jsonmod = %s" % jsonapi.jsonmod
-            else:
-                jsonmsg = ""
             raise ValueError(
-                msg.format(packer=self.packer, unpacker=self.unpacker, e=e, jsonmsg=jsonmsg)
+                f"unpacker '{self.unpacker}' could not handle output from packer"
+                f" '{self.packer}': {e}"
             ) from e
 
         # check datetime support
-        msg = dict(t=utcnow())
+        msg_datetime = dict(t=utcnow())
         try:
-            unpacked = unpack(pack(msg))
-            if isinstance(unpacked['t'], datetime):
+            unpacked = unpack(pack(msg_datetime))
+            if isinstance(unpacked["t"], datetime):
                 raise ValueError("Shouldn't deserialize to datetime")
         except Exception:
             self.pack = lambda o: pack(squash_dates(o))
             self.unpack = lambda s: unpack(s)
 
-    def msg_header(self, msg_type):
+    def msg_header(self, msg_type: str) -> t.Dict[str, t.Any]:
         return msg_header(self.msg_id, msg_type, self.username, self.session)
 
-    def msg(self, msg_type, content=None, parent=None, header=None, metadata=None):
+    def msg(
+        self,
+        msg_type: str,
+        content: t.Optional[t.Dict] = None,
+        parent: t.Optional[t.Dict[str, t.Any]] = None,
+        header: t.Optional[t.Dict[str, t.Any]] = None,
+        metadata: t.Optional[t.Dict[str, t.Any]] = None,
+    ) -> t.Dict[str, t.Any]:
         """Return the nested message dict.
 
         This format is different from what is sent over the wire. The
@@ -573,17 +659,17 @@ class Session(Configurable):
         """
         msg = {}
         header = self.msg_header(msg_type) if header is None else header
-        msg['header'] = header
-        msg['msg_id'] = header['msg_id']
-        msg['msg_type'] = header['msg_type']
-        msg['parent_header'] = {} if parent is None else extract_header(parent)
-        msg['content'] = {} if content is None else content
-        msg['metadata'] = self.metadata.copy()
+        msg["header"] = header
+        msg["msg_id"] = header["msg_id"]
+        msg["msg_type"] = header["msg_type"]
+        msg["parent_header"] = {} if parent is None else extract_header(parent)
+        msg["content"] = {} if content is None else content
+        msg["metadata"] = self.metadata.copy()
         if metadata is not None:
-            msg['metadata'].update(metadata)
+            msg["metadata"].update(metadata)
         return msg
 
-    def sign(self, msg_list):
+    def sign(self, msg_list: t.List) -> bytes:
         """Sign a message with HMAC digest. If no auth, return b''.
 
         Parameters
@@ -592,13 +678,17 @@ class Session(Configurable):
             The [p_header,p_parent,p_content] part of the message list.
         """
         if self.auth is None:
-            return b''
+            return b""
         h = self.auth.copy()
         for m in msg_list:
             h.update(m)
-        return str_to_bytes(h.hexdigest())
+        return h.hexdigest().encode()
 
-    def serialize(self, msg, ident=None):
+    def serialize(
+        self,
+        msg: t.Dict[str, t.Any],
+        ident: t.Optional[t.Union[t.List[bytes], bytes]] = None,
+    ) -> t.List[bytes]:
         """Serialize the message components to bytes.
 
         This is roughly the inverse of deserialize. The serialize/deserialize
@@ -621,7 +711,7 @@ class Session(Configurable):
             In this list, the ``p_*`` entities are the packed or serialized
             versions, so if JSON is used, these are utf8 encoded JSON strings.
         """
-        content = msg.get('content', {})
+        content = msg.get("content", {})
         if content is None:
             content = self.none
         elif isinstance(content, dict):
@@ -631,14 +721,15 @@ class Session(Configurable):
             pass
         elif isinstance(content, str):
             # should be bytes, but JSON often spits out unicode
-            content = content.encode('utf8')
+            content = content.encode("utf8")
         else:
-            raise TypeError("Content incorrect type: %s"%type(content))
+            raise TypeError("Content incorrect type: %s" % type(content))
 
-        real_message = [self.pack(msg['header']),
-                        self.pack(msg['parent_header']),
-                        self.pack(msg['metadata']),
-                        content,
+        real_message = [
+            self.pack(msg["header"]),
+            self.pack(msg["parent_header"]),
+            self.pack(msg["metadata"]),
+            content,
         ]
 
         to_send = []
@@ -657,8 +748,18 @@ class Session(Configurable):
 
         return to_send
 
-    def send(self, stream, msg_or_type, content=None, parent=None, ident=None,
-             buffers=None, track=False, header=None, metadata=None):
+    def send(
+        self,
+        stream: Optional[Union[zmq.sugar.socket.Socket, ZMQStream]],
+        msg_or_type: t.Union[t.Dict[str, t.Any], str],
+        content: t.Optional[t.Dict[str, t.Any]] = None,
+        parent: t.Optional[t.Dict[str, t.Any]] = None,
+        ident: t.Optional[t.Union[bytes, t.List[bytes]]] = None,
+        buffers: t.Optional[t.List[bytes]] = None,
+        track: bool = False,
+        header: t.Optional[t.Dict[str, t.Any]] = None,
+        metadata: t.Optional[t.Dict[str, t.Any]] = None,
+    ) -> t.Optional[t.Dict[str, t.Any]]:
         """Build and send a message via stream or socket.
 
         The message format used by this function internally is as follows:
@@ -710,15 +811,18 @@ class Session(Configurable):
             # We got a Message or message dict, not a msg_type so don't
             # build a new Message.
             msg = msg_or_type
-            buffers = buffers or msg.get('buffers', [])
+            buffers = buffers or msg.get("buffers", [])
         else:
-            msg = self.msg(msg_or_type, content=content, parent=parent,
-                           header=header, metadata=metadata)
-        if self.check_pid and not os.getpid() == self.pid:
-            get_logger().warning("WARNING: attempted to send message from fork\n%s",
-                msg
+            msg = self.msg(
+                msg_or_type,
+                content=content,
+                parent=parent,
+                header=header,
+                metadata=metadata,
             )
-            return
+        if self.check_pid and not os.getpid() == self.pid:
+            get_logger().warning("WARNING: attempted to send message from fork\n%s", msg)
+            return None
         buffers = [] if buffers is None else buffers
         for idx, buf in enumerate(buffers):
             if isinstance(buf, memoryview):
@@ -731,7 +835,7 @@ class Session(Configurable):
                     raise TypeError("Buffer objects must support the buffer protocol.") from e
             # memoryview.contiguous is new in 3.3,
             # just skip the check on Python 2
-            if hasattr(view, 'contiguous') and not view.contiguous:
+            if hasattr(view, "contiguous") and not view.contiguous:
                 # zmq requires memoryviews to be contiguous
                 raise ValueError("Buffer %i (%r) is not contiguous" % (idx, buf))
 
@@ -739,13 +843,13 @@ class Session(Configurable):
             msg = adapt(msg, self.adapt_version)
         to_send = self.serialize(msg, ident)
         to_send.extend(buffers)
-        longest = max([ len(s) for s in to_send ])
-        copy = (longest < self.copy_threshold)
+        longest = max([len(s) for s in to_send])
+        copy = longest < self.copy_threshold
 
-        if buffers and track and not copy:
+        if stream and buffers and track and not copy:
             # only really track when we are doing zero-copy buffers
             tracker = stream.send_multipart(to_send, copy=False, track=True)
-        else:
+        elif stream:
             # use dummy tracker, which will be done immediately
             tracker = DONE
             stream.send_multipart(to_send, copy=copy)
@@ -755,11 +859,18 @@ class Session(Configurable):
             pprint.pprint(to_send)
             pprint.pprint(buffers)
 
-        msg['tracker'] = tracker
+        msg["tracker"] = tracker
 
         return msg
 
-    def send_raw(self, stream, msg_list, flags=0, copy=True, ident=None):
+    def send_raw(
+        self,
+        stream: zmq.sugar.socket.Socket,
+        msg_list: t.List,
+        flags: int = 0,
+        copy: bool = True,
+        ident: t.Optional[t.Union[bytes, t.List[bytes]]] = None,
+    ) -> None:
         """Send a raw message via ident path.
 
         This method is used to send a already serialized message.
@@ -787,7 +898,13 @@ class Session(Configurable):
         to_send.extend(msg_list)
         stream.send_multipart(to_send, flags, copy=copy)
 
-    def recv(self, socket, mode=zmq.NOBLOCK, content=True, copy=True):
+    def recv(
+        self,
+        socket: zmq.sugar.socket.Socket,
+        mode: int = zmq.NOBLOCK,
+        content: bool = True,
+        copy: bool = True,
+    ) -> t.Tuple[t.Optional[t.List[bytes]], t.Optional[t.Dict[str, t.Any]]]:
         """Receive and unpack a message.
 
         Parameters
@@ -809,7 +926,7 @@ class Session(Configurable):
             if e.errno == zmq.EAGAIN:
                 # We can convert EAGAIN to None as we know in this case
                 # recv_multipart won't return None.
-                return None,None
+                return None, None
             else:
                 raise
         # split multipart message into identity list and message dict
@@ -821,7 +938,9 @@ class Session(Configurable):
             # TODO: handle it
             raise e
 
-    def feed_identities(self, msg_list, copy=True):
+    def feed_identities(
+        self, msg_list: t.Union[t.List[bytes], t.List[zmq.Message]], copy: bool = True
+    ) -> t.Tuple[t.List[bytes], t.Union[t.List[bytes], t.List[zmq.Message]]]:
         """Split the identities from the rest of the message.
 
         Feed until DELIM is reached, then return the prefix as idents and
@@ -845,20 +964,22 @@ class Session(Configurable):
             point.
         """
         if copy:
+            msg_list = t.cast(t.List[bytes], msg_list)
             idx = msg_list.index(DELIM)
-            return msg_list[:idx], msg_list[idx+1:]
+            return msg_list[:idx], msg_list[idx + 1 :]  # noqa
         else:
+            msg_list = t.cast(t.List[zmq.Message], msg_list)
             failed = True
-            for idx,m in enumerate(msg_list):
+            for idx, m in enumerate(msg_list):  # noqa
                 if m.bytes == DELIM:
                     failed = False
                     break
             if failed:
                 raise ValueError("DELIM not in msg_list")
-            idents, msg_list = msg_list[:idx], msg_list[idx+1:]
-            return [m.bytes for m in idents], msg_list
+            idents, msg_list = msg_list[:idx], msg_list[idx + 1 :]  # noqa
+            return [bytes(m.bytes) for m in idents], msg_list
 
-    def _add_digest(self, signature):
+    def _add_digest(self, signature: bytes) -> None:
         """add a digest to history to protect against replay attacks"""
         if self.digest_history_size == 0:
             # no history, never add digests
@@ -869,7 +990,7 @@ class Session(Configurable):
             # threshold reached, cull 10%
             self._cull_digest_history()
 
-    def _cull_digest_history(self):
+    def _cull_digest_history(self) -> None:
         """cull the digest history
 
         Removes a randomly selected 10% of the digest history
@@ -879,10 +1000,15 @@ class Session(Configurable):
         if n_to_cull >= current:
             self.digest_history = set()
             return
-        to_cull = random.sample(self.digest_history, n_to_cull)
+        to_cull = random.sample(tuple(sorted(self.digest_history)), n_to_cull)
         self.digest_history.difference_update(to_cull)
 
-    def deserialize(self, msg_list, content=True, copy=True):
+    def deserialize(
+        self,
+        msg_list: t.Union[t.List[bytes], t.List[zmq.Message]],
+        content: bool = True,
+        copy: bool = True,
+    ) -> t.Dict[str, t.Any]:
         """Unserialize a msg_list to a nested message dict.
 
         This is roughly the inverse of serialize. The serialize/deserialize
@@ -911,8 +1037,11 @@ class Session(Configurable):
         message = {}
         if not copy:
             # pyzmq didn't copy the first parts of the message, so we'll do it
-            for i in range(minlen):
-                msg_list[i] = msg_list[i].bytes
+            msg_list = t.cast(t.List[zmq.Message], msg_list)
+            msg_list_beginning = [bytes(msg.bytes) for msg in msg_list[:minlen]]
+            msg_list = t.cast(t.List[bytes], msg_list)
+            msg_list = msg_list_beginning + msg_list[minlen:]
+        msg_list = t.cast(t.List[bytes], msg_list)
         if self.auth is not None:
             signature = msg_list[0]
             if not signature:
@@ -926,47 +1055,31 @@ class Session(Configurable):
             if not compare_digest(signature, check):
                 raise ValueError("Invalid Signature: %r" % signature)
         if not len(msg_list) >= minlen:
-            raise TypeError("malformed message, must have at least %i elements"%minlen)
+            raise TypeError("malformed message, must have at least %i elements" % minlen)
         header = self.unpack(msg_list[1])
-        message['header'] = extract_dates(header)
-        message['msg_id'] = header['msg_id']
-        message['msg_type'] = header['msg_type']
-        message['parent_header'] = extract_dates(self.unpack(msg_list[2]))
-        message['metadata'] = self.unpack(msg_list[3])
+        message["header"] = extract_dates(header)
+        message["msg_id"] = header["msg_id"]
+        message["msg_type"] = header["msg_type"]
+        message["parent_header"] = extract_dates(self.unpack(msg_list[2]))
+        message["metadata"] = self.unpack(msg_list[3])
         if content:
-            message['content'] = self.unpack(msg_list[4])
+            message["content"] = self.unpack(msg_list[4])
         else:
-            message['content'] = msg_list[4]
+            message["content"] = msg_list[4]
         buffers = [memoryview(b) for b in msg_list[5:]]
         if buffers and buffers[0].shape is None:
             # force copy to workaround pyzmq #646
-            buffers = [memoryview(b.bytes) for b in msg_list[5:]]
-        message['buffers'] = buffers
+            msg_list = t.cast(t.List[zmq.Message], msg_list)
+            buffers = [memoryview(bytes(b.bytes)) for b in msg_list[5:]]
+        message["buffers"] = buffers
         if self.debug:
             pprint.pprint(message)
         # adapt to the current version
         return adapt(message)
 
-    def unserialize(self, *args, **kwargs):
+    def unserialize(self, *args: Any, **kwargs: Any) -> t.Dict[str, t.Any]:
         warnings.warn(
             "Session.unserialize is deprecated. Use Session.deserialize.",
             DeprecationWarning,
         )
         return self.deserialize(*args, **kwargs)
-
-
-def test_msg2obj():
-    am = dict(x=1)
-    ao = Message(am)
-    assert ao.x == am['x']
-
-    am['y'] = dict(z=1)
-    ao = Message(am)
-    assert ao.y.z == am['y']['z']
-
-    k1, k2 = 'y', 'z'
-    assert ao[k1][k2] == am[k1][k2]
-
-    am2 = dict(ao)
-    assert am['x'] == am2['x']
-    assert am['y']['z'] == am2['y']['z']
